##
##    Title:    approxrad
##    Created:  Friday, March 14 2008
##    Author:  Itnuit Janovitz-Freireich <ijanovi2@ncsu.edu>        
##    Contributions of:
##              
## Description: 
##
##
## Version last modified 02/2009
## ----------------------------------------------------------------------
##
##  The approxrad Maple file gathers some functions for computing radicals 
##      and approximate radicals of polynomial systems.
##
##  
## 
##  You can use it with a simple : 
##      read "approxrad.mpl";
##
##  The file is divided in the following parts:
##
##  1) Matrix manipulations: Basic functions to work with polynomials
##         are developed. 
##
##  2) Polynomial manipulations.
##
##  3) Multiplication, trace and moment matrices.
## 
## ----------------------------------------------------------------------

########################################################################################
########################################################################################

with(linalg): with(combinat,choose):with(LinearAlgebra):
__ := NULL: addtohelp := proc() global __; __ := __, args: end:

print("You need to load multires and chardin.");


#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXX MATRIX MANIPULATIONS XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#


##########################################################################################
##########################################################################################

## The following program reorders the columns in the input matrix
## so that they'll be in the same order as the input monomial lists. 
## The imput is a matrix, its index of rows, index of cols, 
## index we want for the rows (can have less elements than the row), 
## index we want for the columns. The output are three matrices, 
## corresponding to the reordered matrix, the reordered rows, the reordered columns.

ReorderMatrix:=proc(Mat,row,col,basr,basc)
  local rowr,colr,i,j,Br;
  rowr:=convert(row,Matrix);
  colr:=convert(col,Matrix);
  Br:=Mat;
  for i from 1 to nops(basr) do
    for j from 1 to nops(row) do
      if rowr[1,j]=basr[i] then
        Br:=RowOperation(Br,[i,j]);
        rowr:=ColumnOperation(rowr,[i,j]);
      fi;
    od;
  od;
  for i from 1 to nops(basc) do
    for j from 1 to nops(col) do
      if colr[1,j]=basc[i] then
        Br:=ColumnOperation(Br,[i,j]);
        colr:=ColumnOperation(colr,[i,j]);
      fi;
    od;
  od;
return Br,rowr,colr;
end:

##########################################################################################
##########################################################################################

## The following two programs take the matrix Mat
## and replaces its r-th row/col by list.
## Note that list needs to be a column vector in the second case

ReplaceRow:=proc(Mat,r,list)
  local C,A;
  C:=coldim(Mat);
  A:=SubMatrix(Mat,[seq(i,i=1..r-1)],[1..C]);
  A:=convert(A,matrix);
  A:=stackmatrix(A,list);
  A:=stackmatrix(A,SubMatrix(Mat,[seq(i,i=r+1..rowdim(Mat))],[1..C]));
return(convert(A, Matrix));
end:

ReplaceCol:=proc(Mat,r,list)
  local C,A,B;
  C:=rowdim(Mat);
  A:=SubMatrix(Mat,[1..C],[seq(i,i=1..r-1)]);
  B:=convert(SubMatrix(Mat,[1..C],[seq(i,i=r+1..coldim(Mat))]),matrix);
  A:=convert(A,matrix);
  A:=augment(A,list);
  A:=augment(A,B);
return(convert(A,Matrix));
end:

##########################################################################################
##########################################################################################

## The following program reorders a matrix, indexed by row and col,
## and adds zero columns and rows so that it is now expressed in terms of basr and basc.
## The imput is a bezout matrix, its index of rows (list), index of cols (list), 
## index we want for the rows basr (list) and
## index we want for the columns basc (list). basr and basc must contain row and col.
## The output is a reordered and expanded matrix, ordered by basr, basc.

ReorderExpandMatrix:=proc(Mat,row,col,basr,basc)
  local i,j,Br,Bc;
  Br:=ZeroMatrix(nops(basr),nops(col));
  for i from 1 to nops(basr) do
    for j from 1 to nops(row) do
        if row[j]=basr[i] then
        Br:=ReplaceRow(Br,i,Row(Mat,j));
        fi;
    od;
  od;
  Bc:=ZeroMatrix(nops(basr),nops(basc));
  for i from 1 to nops(basc) do
    for j from 1 to nops(col) do
      if col[j]=basc[i] then
        Bc:=ReplaceCol(Bc,i,Column(Br,j));
      fi;
    od;
  od;
return Bc;
end:


########################################################################################
########################################################################################

## The following program does Gaussian Elimination with Complete Pivoting
## on the matrix M, with rows indexed by Basr and columns indexed by Basc.
## The output is the list of new indexes of the rows and columns and the obtained 
## eliminated matrix U. 

CompletePivoting:=proc(M,Basr,Basc)
  local P, Pp,L,i,j,p,k,U;
  P:=[Basr,Basc];
  Pp:=[[],[]];
  U:=M;
  if ColumnDimension(M) <> RowDimension(M) then return "Not a square Matrix.";fi;
  for k from 1 to ColumnDimension(M)-1 do
    L:=[[seq(m,m=k..ColumnDimension(M))],[seq(m,m=k..RowDimension(M))]];
    for i in L[1] do
      for j in L[2] do
        if U[i,j]=max(seq(seq(U[s,k],s=k..ColumnDimension(M)),k=k..RowDimension(M))) and U[i,j]<>0
        then
          p:=[i,j];
          U:=RowOperation(U,[k,i]);
          U:=ColumnOperation(U,[k,j]);
          U:=Pivot(U,k,k,[k..ColumnDimension(M)]);
          P:=[convert(P[1],Matrix),convert(P[2],Matrix)];
          P:=[ColumnOperation(P[1],[k,i]),ColumnOperation(P[2],[k,j])];
        fi;
      od;
    od;
  od;
  return([P,U]);
  end:

########################################################################################
########################################################################################

## The following program does N steps of Gaussian Elimination with Complete Pivoting
## on the matrix M, with rows indexed by Basr and columns indexed by Basc.
## The output is the list of new indexes of the rows and columns and the obtained 
## eliminated matrix U. 


PartialCompletePivoting:=proc(M,Basr,Basc,N)
  local P, Pp,L,i,j,p,k,U;
  P:=[Basr,Basc];
  Pp:=[[],[]];
  U:=M;
  if ColumnDimension(M) <> RowDimension(M) then return "Not a square Matrix.";fi;
  for k from 1 to N do
    L:=[[seq(m,m=k..ColumnDimension(M))],[seq(m,m=k..RowDimension(M))]];
    for i in L[1] do
      for j in L[2] do
        if   
          U[i,j]=max(seq(seq(U[s,k],s=k..ColumnDimension(M)),k=k..RowDimension(M))) and U[i,j]<>0
        then 
          p:=[i,j]; 
          U:=RowOperation(U,[k,i]);
          U:=ColumnOperation(U,[k,j]);
          U:=Pivot(U,k,k,[k..ColumnDimension(M)]);
          P:=[convert(P[1],Matrix),convert(P[2],Matrix)];
          P:=[ColumnOperation(P[1],[k,i]),ColumnOperation(P[2],[k,j])];
        fi;
      od;
    od;
  od;
return([P,U]);
end:
#########################################################################################
#########################################################################################



#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXX POLYNOMIAl MANIPULATIONS XXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#


############################################################################################
############################################################################################

## The following program extracts the coefficient of the monomial in the polynomial poly, 
## in the variables vars ({x1, x2}, for example).

coeffmu := proc(poly,monomial,vars)
  local c, i;
  c := coeffs(poly, vars, 'T');
  T;
  for i from 1 to nops([T]) do
    if T[i] = monomial then
      return c[i];
    fi;
  od; 
end:
#########################################################################################
#########################################################################################







#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXX MULTIPLICATION             XXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXX TRACE             MATRICES XXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXX MOMENT                     XXXXXXXXXXXXXXXXXXXXXXXXXXX#
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX#


#########################################################################################
#########################################################################################

## The following program returns the multiplication matrices for x1 or  x2  
##(these need to be chosen as var) of a system of 2 (dense) polynomials
## in the variables x1 and x2 with respect to the smallest degree basis. 
## The output is a list with the basis and the multiplication matrix

Multmatrices:=proc(poly1, poly2, var, dima)
local i,m,n, h1, polyh1, h2, polyh2, Cx, dim, C, list, ol, l, k, j, s, b, Crx, mr, l1;

##Construct the homogeneous version of the polynomials:
h1:=[];
for i from 1 to nops(poly1) do
  m:=max(seq(degree(op(poly1)[j]),j=1..nops(poly1)));
  n:=degree(op(poly1)[i]);
  h1:=[op(h1),op(poly1)[i]*z^(m-n)];
od;
polyh1:=add(h1[i],i=1..nops(h1));
h2:=[];
for i from 1 to nops(poly2) do
  m:=max(seq(degree(op(poly2)[j]),j=1..nops(poly2)));
  n:=degree(op(poly2)[i]);
  h2:=[op(h2),op(poly2)[i]*z^(m-n)];
od;
polyh2:=add(h2[i],i=1..nops(h2));

##Determine dimension
if nargs=4 then
  dim:=dima;
else
  dim:=degree(polyh1)*degree(polyh2);
fi;

##Construct the Chardin subresultant for var:
Cx:=Chardin([polyh1,polyh2,var],[x1,x2,z],(degree(polyh1)-1)+(degree(polyh2)-1)+1);
rank(Cx[1]);
m:=map(X->subs(z=1,X), Cx[3]);

##Determine the elements that will be in the basis
list:=[1];
for i from 1 to dim do
  ol:=[];
  l:=listmonhomdegnu([x1,x2],i);
  for k from 1 to i do
    for j from 1 to nops(l) do
      if degree(l[j],x1)=k then
      ol:=[op(ol),l[j]];
      fi;
    od;
  od;
  for s from 1 to nops(l) do
    if degree(l[s],x1)=0 then
    ol:=[op(ol),l[s]];
    fi;
  od;
  list:=[op(list),op(ol)];
od;
b:=[];
for i from 1 to nops(list) do
  for j from 1 to coldim(m) do
    if list[i]=m[1,j] then
      b:=[op(b),list[i]];
      if nops(b)=dim then break; fi;
    fi;
   if nops(b)=dim then break; fi;
   od;
   if nops(b)=dim then break; fi;
od;

##We now need to reorder the Chardin Matrix to get the elements we want at the end:
Crx:=convert(Transpose(Cx[1]),Matrix);
m:=convert(m,Matrix);
mr:=m;
nops(b);
for i from 1 to nops(b) do
  b[i];i;
  for j from 1 to coldim(m) do
    if mr[1,j]=b[i] then
      Crx:=RowOperation(Crx,[coldim(m)-nops(b)+i,j]);
      Crx:=ColumnOperation(Crx,[coldim(m)-nops(b)+i,j]);
      mr:=ColumnOperation(mr,[coldim(m)-nops(b)+i,j]);
    fi;
  od;
od;

##We can now take the schur complement of the chardin matrix:
return [b,transpose(lschur(Crx,coldim(Cx[1])-dim))];
end:
###########################################################################################
###########################################################################################


## The following program computes the matrix of traces and the minimum 
## degree basis of a system given two polynomials in the variables x1 and x2 using multiplication matrices.
## The output is a list with the basis and the matrix.

Multtraces:=proc(poly1,poly2)
local a,bas, mx1,mx2,i, j, p, Mtr;
a:=Multmatrices(poly1,poly2,x1);
bas:=a[1];
mx1:=simplify(convert(a[2],Matrix));
mx2:=simplify(convert(Multmatrices(poly1,poly2,x2)[2],Matrix));
for i from 1 to nops(bas) do
  for j from 1 to nops(bas) do
p[i,j]:=simplify(trace(simplify(Multiply(simplify(MatrixPower(mx1,degree(bas[i]*bas[j],x1))), 
           simplify(MatrixPower(mx2,degree(bas[i]*bas[j],x2)))))));
  od;
od;
Mtr:=Matrix(nops(bas),(i,j)->p[i,j]); 
print([bas, rank(Mtr),rank(evalf(Mtr)),evalf(SingularValues(Mtr)),evalf(SingularValues(evalf(Mtr)))]);
return [bas, Mtr,rank(Mtr),evalf(Mtr),rank(evalf(Mtr))];
end:
##########################################################################################
##########################################################################################


